from collections.abc import Callable, Iterator
from functools import reduce

from edgeql_qb.expression import Expression
from edgeql_qb.render.expression import render_expression
from edgeql_qb.render.tools import combine_many_renderers, join_renderers
from edgeql_qb.render.types import RenderedQuery


def render_with_expressions(
        expressions: list[RenderedQuery],
) -> Callable[[RenderedQuery], RenderedQuery]:
    def inner(rendered_with: RenderedQuery) -> RenderedQuery:
        return (
            combine_many_renderers(
                rendered_with,
                reduce(join_renderers(', '), expressions),
            )
            if expressions
            else rendered_with
        )
    return inner


def render_with_module(module: str | None = None) -> Callable[[RenderedQuery], RenderedQuery]:
    def inner(rendered_with: RenderedQuery) -> RenderedQuery:
        return (
            combine_many_renderers(
                rendered_with,
                RenderedQuery(f'module {module}'),
            )
            if module
            else rendered_with
        )
    return inner


def render_with_separator(
    expressions: list[RenderedQuery],
    module: str | None = None,
) -> Callable[[RenderedQuery], RenderedQuery]:
    """Put comma between module and expressions if both are presented."""
    def inner(rendered_with: RenderedQuery) -> RenderedQuery:
        comma = RenderedQuery(', ')
        return (
            combine_many_renderers(rendered_with, comma)
            if (module and expressions)
            else rendered_with
        )
    return inner


def render_with_ending(
    expressions: list[RenderedQuery],
    module: str | None = None,
) -> Callable[[RenderedQuery], RenderedQuery]:
    """Put space on the end of expressions if any module or expressions presented."""
    def inner(rendered_with: RenderedQuery) -> RenderedQuery:
        space = RenderedQuery(' ')
        empty = RenderedQuery()
        return (
            combine_many_renderers(rendered_with, space)
            if module or expressions else space
            if module and expressions else empty
        )
    return inner


def render_with_expression(
        with_aliases: tuple[Expression, ...],
        generator: Iterator[int],
        module: str | None = None,
) -> RenderedQuery:
    renderers = [
        render_expression(alias.to_infix_notation(), 'with', generator)
        for alias in with_aliases
    ]
    return (
        RenderedQuery('with ')
        .map(render_with_module(module))
        .map(render_with_separator(renderers, module))
        .map(render_with_expressions(renderers))
        .map(render_with_ending(renderers, module))
    )



from datetime import date, datetime, time, timedelta
from decimal import Decimal
from functools import singledispatch
from typing import Any

from edgeql_qb.frozendict import FrozenDict
from edgeql_qb.render.types import RenderedQuery
from edgeql_qb.types import GenericHolder


@singledispatch
def render_query_literal(value: Any, name: str) -> RenderedQuery:
    if isinstance(value, str | bool | bytes):
        # singledispatch not working for unions
        return RenderedQuery(
            f'<{value.__class__.__name__}>${name}',
            FrozenDict({name: value}),
        )
    return RenderedQuery(f'${name}', FrozenDict({name: value}))  # pragma: no cover


@render_query_literal.register
def _(value: GenericHolder, name: str) -> RenderedQuery:  # type: ignore[type-arg]
    return RenderedQuery(f'<{value.edgeql_name}>${name}', FrozenDict({name: value.value}))


@render_query_literal.register
def _(value: datetime, name: str) -> RenderedQuery:
    if value.tzinfo is None:
        return RenderedQuery(f'<cal::local_datetime>${name}', FrozenDict({name: value}))
    return RenderedQuery(f'<{value.__class__.__name__}>${name}', FrozenDict({name: value}))


@render_query_literal.register
def _(value: date, name: str) -> RenderedQuery:
    return RenderedQuery(f'<cal::local_date>${name}', FrozenDict({name: value}))


@render_query_literal.register
def _(value: time, name: str) -> RenderedQuery:
    return RenderedQuery(f'<cal::local_time>${name}', FrozenDict({name: value}))


@render_query_literal.register
def _(value: timedelta, name: str) -> RenderedQuery:
    return RenderedQuery(f'<duration>${name}', FrozenDict({name: value}))


@render_query_literal.register
def _(value: Decimal, name: str) -> RenderedQuery:
    return RenderedQuery(f'<decimal>${name}', FrozenDict({name: value}))



from collections.abc import Iterator
from functools import reduce, singledispatch

from edgeql_qb.expression import (
    AnyExpression,
    Column,
    Expression,
    QueryLiteral,
)
from edgeql_qb.func import FuncInvocation
from edgeql_qb.operators import Alias, Node
from edgeql_qb.render.func import render_function
from edgeql_qb.render.query_literal import render_query_literal
from edgeql_qb.render.tools import (
    combine_many_renderers,
    combine_renderers,
    join_renderers,
    linearize_filter_left,
    render_binary_node,
)
from edgeql_qb.render.types import RenderedQuery
from edgeql_qb.types import unsafe_text


def render_filters(filters: tuple[Expression, ...], generator: Iterator[int]) -> RenderedQuery:
    conditions = [
        render_condition(filter_.to_infix_notation(), generator)
        for filter_ in filters
    ]
    return combine_renderers(
        RenderedQuery(' filter '),
        reduce(join_renderers(' and '), conditions),
    )


def render_conditions(filters: tuple[Expression, ...], generator: Iterator[int]) -> RenderedQuery:
    return render_filters(filters, generator) if filters else RenderedQuery()


@singledispatch
def render_condition(expression: AnyExpression, generator: Iterator[int]) -> RenderedQuery:
    raise NotImplementedError(f'{expression!r} is not supported')  # pragma: no cover


@render_condition.register
def _(expression: Alias, generator: Iterator[int]) -> RenderedQuery:
    return RenderedQuery(expression.name)


@render_condition.register
def _(expression: FuncInvocation, generator: Iterator[int]) -> RenderedQuery:
    func = expression.func
    arg_renderers = [
        render_condition(arg, generator)
        for arg in expression.args
    ]
    return render_function(func, arg_renderers)


@render_condition.register
def _(expression: Node, generator: Iterator[int]) -> RenderedQuery:
    if expression.right is None:
        return combine_many_renderers(
            RenderedQuery(expression.op),
            render_condition(expression.left, generator),
        )
    return render_binary_node(
        left=render_condition(expression.left, generator),
        right=render_condition(expression.right, generator),
        expression=expression,
    )


@render_condition.register
def _(expression: QueryLiteral, generator: Iterator[int]) -> RenderedQuery:
    index = next(generator)
    name = f'filter_{index}'
    return render_query_literal(expression.value, name)


@render_condition.register
def _(expression: unsafe_text, generator: Iterator[int]) -> RenderedQuery:
    return RenderedQuery(expression)


@render_condition.register
def _(expression: Column, generator: Iterator[int]) -> RenderedQuery:
    columns = linearize_filter_left(expression)
    dot_names = '.'.join(c.column_name for c in columns)
    return RenderedQuery(f'.{dot_names}')



from typing import Generic, TypeVar

T = TypeVar('T')


class GenericHolder(Generic[T]):
    __match_args__ = 'edgeql_name', 'value'

    edgeql_name: str

    def __init__(self, value: T):
        self.value = value

    def __repr__(self) -> str:
        return f'<{self.edgeql_name}>{self.value}'


class int16(GenericHolder[int]):
    edgeql_name = 'int16'


class int32(GenericHolder[int]):
    edgeql_name = 'int32'


class int64(GenericHolder[int]):
    edgeql_name = 'int64'


class bigint(GenericHolder[int]):
    edgeql_name = 'bigint'


class float32(GenericHolder[float]):
    edgeql_name = 'float32'


class float64(GenericHolder[float]):
    edgeql_name = 'float64'


class unsafe_text(str):
    """Wrapper for rendering a text node as is.

    WARNING: Don't pass unvalidated user input to this wrapper.
    Otherwise, you may be the first victim of EdgeQL Injection attack.
    There's no honor in this.
    """



from collections.abc import Iterator
from dataclasses import dataclass, field, replace
from itertools import count
from typing import Any

from edgeql_qb.expression import (
    BaseModel,
    Column,
    Columns,
    Expression,
    SelectExpressions,
    SubQuery,
    UnlessConflict,
    UpdateSubQuery,
)
from edgeql_qb.func import FuncInvocation
from edgeql_qb.operators import BinaryOp, SortedExpression, UnaryOp
from edgeql_qb.render.condition import render_conditions
from edgeql_qb.render.delete import render_delete
from edgeql_qb.render.group import (
    render_group,
    render_group_by_expressions,
    render_using_expressions,
)
from edgeql_qb.render.insert import render_insert, render_unless_conflict
from edgeql_qb.render.insert import render_values as render_insert_values
from edgeql_qb.render.order_by import render_order_by
from edgeql_qb.render.pagination import render_limit, render_offset
from edgeql_qb.render.select import render_select
from edgeql_qb.render.tools import combine_many_renderers
from edgeql_qb.render.types import RenderedQuery
from edgeql_qb.render.update import render_update
from edgeql_qb.render.update import render_values as render_update_values
from edgeql_qb.render.with_expr import render_with_expression
from edgeql_qb.types import unsafe_text


@dataclass(slots=True, frozen=True)
class EdgeDBModel(BaseModel):
    c: Columns = field(default_factory=Columns)

    def select(self, *selectables: SelectExpressions, **kwargs: SubQuery) -> 'SelectQuery':
        select_args = tuple(Expression(sel) for sel in selectables)
        select_kwargs = tuple(Expression(v.label(k)) for k, v in kwargs.items())
        return SelectQuery(
            _model=self,
            _select=(*select_args, *select_kwargs),
        )

    def group(self, *selectables: SelectExpressions) -> 'GroupQuery':
        return GroupQuery(
            _model=self,
            _select=tuple(Expression(sel) for sel in selectables),
        )

    @property
    def delete(self) -> 'DeleteQuery':
        return DeleteQuery(self)

    @property
    def insert(self) -> 'InsertQuery':
        return InsertQuery(self)

    @property
    def update(self) -> 'UpdateQuery':
        return UpdateQuery(self)


@dataclass(slots=True, frozen=True)
class SelectQuery(SubQuery):
    _model: EdgeDBModel
    _select: tuple[Expression, ...] = field(default_factory=tuple)
    _select_from_query: SubQuery | None = None
    _with_aliases: tuple[Expression, ...] = field(default_factory=tuple)
    _filters: tuple[Expression, ...] = field(default_factory=tuple)
    _ordered_by: tuple[Expression, ...] = field(default_factory=tuple)
    _limit_val: int | unsafe_text | FuncInvocation | None = None
    _offset_val: int | unsafe_text | FuncInvocation | None = None

    def select_from(self, query: SubQuery) -> 'SelectQuery':
        return replace(self, _select_from_query=query)

    def where(self, compared: BinaryOp | UnaryOp | FuncInvocation) -> 'SelectQuery':
        return replace(self, _filters=(*self._filters, Expression(compared)))

    def with_(self, *with_aliases: BinaryOp) -> 'SelectQuery':
        expressions = tuple(Expression(exp) for exp in with_aliases)
        return replace(self, _with_aliases=expressions)

    def order_by(
        self,
        *columns: SortedExpression | Column | UnaryOp | FuncInvocation,
    ) -> 'SelectQuery':
        new_expressions = [Expression(exp) for exp in columns]
        return replace(self, _ordered_by=(*self._ordered_by, *new_expressions))

    def limit(self, value: int | FuncInvocation | unsafe_text) -> 'SelectQuery':
        return replace(self, _limit_val=value)

    @property
    def limit1(self) -> 'SelectQuery':
        return replace(self, _limit_val=unsafe_text('1'))

    def offset(self, value: int | FuncInvocation | unsafe_text) -> 'SelectQuery':
        return replace(self, _offset_val=value)

    def build(self, generator: Iterator[int] | None = None) -> RenderedQuery:
        gen = generator or count()
        rendered_with = render_with_expression(self._with_aliases, gen, self._model.module)
        rendered_select = render_select(
            self._model.name,
            self._select,
            gen,
            self._select_from_query,
        )
        rendered_filters = render_conditions(self._filters, gen)
        rendered_order_by = render_order_by(self._ordered_by, gen)
        rendered_offset = render_offset(self._offset_val, gen)
        rendered_limit = render_limit(self._limit_val, gen)
        return combine_many_renderers(
            rendered_with,
            rendered_select,
            rendered_filters,
            rendered_order_by,
            rendered_offset,
            rendered_limit,
        )


@dataclass(slots=True, frozen=True)
class GroupQuery:
    _model: EdgeDBModel
    _select: tuple[Expression, ...] = field(default_factory=tuple)
    _with_aliases: tuple[Expression, ...] = field(default_factory=tuple)
    _group_by: tuple[Column | BinaryOp, ...] = field(default_factory=tuple)
    _using_expressions: tuple[Expression, ...] = field(default_factory=tuple)

    def with_(self, *with_aliases: BinaryOp) -> 'GroupQuery':
        expressions = tuple(Expression(exp) for exp in with_aliases)
        return replace(self, _with_aliases=expressions)

    def using(self, *using_expressions: BinaryOp) -> 'GroupQuery':
        expressions = tuple(Expression(exp) for exp in using_expressions)
        return replace(self, _using_expressions=expressions)

    def by(self, *group_by: Column | BinaryOp) -> 'GroupQuery':
        return replace(self, _group_by=group_by)

    def build(self, generator: Iterator[int] | None = None) -> RenderedQuery:
        gen = generator or count()
        rendered_with = render_with_expression(self._with_aliases, gen, self._model.module)
        rendered_group = render_group(self._model.name, self._select, gen)
        rendered_using = render_using_expressions(self._using_expressions, gen)
        rendered_group_by = render_group_by_expressions(self._group_by)
        return combine_many_renderers(
            rendered_with,
            rendered_group,
            rendered_using,
            rendered_group_by,
        )


@dataclass(slots=True, frozen=True)
class DeleteQuery:
    _model: EdgeDBModel
    _with_aliases: tuple[Expression, ...] = field(default_factory=tuple)
    _filters: tuple[Expression, ...] = field(default_factory=tuple)
    _ordered_by: tuple[Expression, ...] = field(default_factory=tuple)
    _limit_val: int | unsafe_text | FuncInvocation | None = None
    _offset_val: int | unsafe_text | FuncInvocation | None = None

    def where(self, compared: BinaryOp | UnaryOp) -> 'DeleteQuery':
        return replace(self, _filters=(*self._filters, Expression(compared)))

    def with_(self, *with_aliases: BinaryOp) -> 'DeleteQuery':
        expressions = tuple(Expression(exp) for exp in with_aliases)
        return replace(self, _with_aliases=expressions)

    def order_by(
        self,
        *columns: SortedExpression | Column | UnaryOp | FuncInvocation,
    ) -> 'DeleteQuery':
        new_expressions = [Expression(exp) for exp in columns]
        return replace(self, _ordered_by=(*self._ordered_by, *new_expressions))

    def limit(self, value: int | FuncInvocation | unsafe_text) -> 'DeleteQuery':
        return replace(self, _limit_val=value)

    def offset(self, value: int | FuncInvocation | unsafe_text) -> 'DeleteQuery':
        return replace(self, _offset_val=value)

    def build(self, generator: Iterator[int] | None = None) -> RenderedQuery:
        gen = generator or count()
        rendered_with = render_with_expression(self._with_aliases, gen, self._model.module)
        rendered_delete = render_delete(self._model.name)
        rendered_filters = render_conditions(self._filters, gen)
        rendered_order_by = render_order_by(self._ordered_by, gen)
        rendered_offset = render_offset(self._offset_val, gen)
        rendered_limit = render_limit(self._limit_val, gen)
        return combine_many_renderers(
            rendered_with,
            rendered_delete,
            rendered_filters,
            rendered_order_by,
            rendered_offset,
            rendered_limit,
        )


@dataclass(slots=True, frozen=True)
class InsertQuery(SubQuery):
    _model: EdgeDBModel
    _with_aliases: tuple[Expression, ...] = field(default_factory=tuple)
    _values_to_insert: list[Expression] = field(default_factory=list)
    _unless_conflict_value: UnlessConflict | None = None

    def values(self, **to_insert: Any) -> 'InsertQuery':
        assert to_insert
        values_to_insert = [
            Expression(BinaryOp(':=', Column(name), exp))
            for name, exp in to_insert.items()
        ]
        return replace(self, _values_to_insert=values_to_insert)

    def with_(self, *with_aliases: BinaryOp) -> 'InsertQuery':
        expressions = tuple(Expression(exp) for exp in with_aliases)
        return replace(self, _with_aliases=expressions)

    def unless_conflict(
        self,
        on: tuple[Column, ...] | Column | None = None,
        else_: UpdateSubQuery | EdgeDBModel | None = None,
    ) -> 'InsertQuery':
        return replace(self, _unless_conflict_value=UnlessConflict(on=on, else_=else_))

    def build(self, generator: Iterator[int] | None = None) -> RenderedQuery:
        assert self._values_to_insert
        gen = generator or count()
        rendered_with = render_with_expression(self._with_aliases, gen, self._model.module)
        rendered_insert = render_insert(self._model.name)
        rendered_values = render_insert_values(self._values_to_insert, gen)
        rendered_conflicts = render_unless_conflict(self._unless_conflict_value, gen)
        return combine_many_renderers(
            rendered_with,
            rendered_insert,
            rendered_values,
            rendered_conflicts,
        )


@dataclass(slots=True, frozen=True)
class UpdateQuery(UpdateSubQuery):
    _model: EdgeDBModel
    _with_aliases: tuple[Expression, ...] = field(default_factory=tuple)
    _values_to_update: tuple[Expression, ...] = field(default_factory=tuple)
    _filters: tuple[Expression, ...] = field(default_factory=tuple)

    def where(self, compared: BinaryOp | UnaryOp) -> 'UpdateQuery':
        return replace(self, _filters=(*self._filters, Expression(compared)))

    def with_(self, *with_aliases: BinaryOp) -> 'UpdateQuery':
        expressions = tuple(Expression(exp) for exp in with_aliases)
        return replace(self, _with_aliases=expressions)

    def values(self, **to_update: Any) -> 'UpdateQuery':
        assert to_update
        values_to_update = tuple(
            Expression(BinaryOp(':=', Column(name), exp))
            for name, exp in to_update.items()
        )
        return replace(self, _values_to_update=values_to_update)

    def build(self, generator: Iterator[int] | None = None) -> RenderedQuery:
        assert self._values_to_update
        gen = generator or count()
        rendered_with = render_with_expression(self._with_aliases, gen, self._model.module)
        rendered_insert = render_update(self._model.name)
        rendered_filters = render_conditions(self._filters, gen)
        rendered_values = render_update_values(self._values_to_update, gen)
        return combine_many_renderers(
            rendered_with,
            rendered_insert,
            rendered_filters,
            rendered_values,
        )



from collections.abc import Iterator
from functools import reduce, singledispatch
from typing import Any

from edgeql_qb.expression import Column, Expression
from edgeql_qb.operators import Alias, BinaryOp
from edgeql_qb.render.expression import render_expression
from edgeql_qb.render.select import render_select_columns
from edgeql_qb.render.tools import combine_renderers, join_renderers
from edgeql_qb.render.types import RenderedQuery


def render_group(
        model_name: str,
        select: tuple[Expression, ...],
        generator: Iterator[int],
) -> RenderedQuery:
    return (
        RenderedQuery(f'group {model_name}')
        .map(
            lambda r: (
                combine_renderers(r, render_select_columns(select, generator))
                if select
                else r
            ),
        )
    )


def render_using(using: tuple[Expression, ...], generator: Iterator[int]) -> RenderedQuery:
    using_expressions = [
        render_expression(use.to_infix_notation(), 'using', generator)
        for use in using
    ]
    return combine_renderers(
        RenderedQuery(' using '),
        reduce(join_renderers(', '), using_expressions),
    )


def render_using_expressions(
        using: tuple[Expression, ...],
        generator: Iterator[int],
) -> RenderedQuery:
    return render_using(using, generator) if using else RenderedQuery()


@singledispatch
def render_group_by(group_by: Any) -> RenderedQuery:
    raise NotImplementedError(f'{group_by!r} is not supported')  # pragma: no cover


@render_group_by.register
def _(group_by: Column) -> RenderedQuery:
    return RenderedQuery(f'.{group_by.column_name}')


@render_group_by.register
def _(group_by: BinaryOp) -> RenderedQuery:
    assert isinstance(group_by.left, Alias)
    return RenderedQuery(group_by.left.name)


def render_group_by_expressions(group_by: tuple[Column | BinaryOp, ...]) -> RenderedQuery:
    renderers = map(render_group_by, group_by)
    return combine_renderers(
        RenderedQuery(' by '),
        reduce(join_renderers(', '), renderers),
    )



from __future__ import annotations

from collections.abc import Iterator, Mapping
from functools import reduce
from typing import Any


class FrozenDict(Mapping[str, Any]):
    """Python hashable mapping type.

    Python 3.10 has no hashable mapping type, so I steel this class from here:
    https://stackoverflow.com/a/2704866.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._d = dict(*args, **kwargs)

    def __iter__(self) -> Iterator[str]:
        return iter(self._d)

    def __len__(self) -> int:
        return len(self._d)

    def __getitem__(self, key: str) -> Any:
        return self._d[key]

    def __or__(self, other: FrozenDict) -> FrozenDict:
        new = dict(self)
        new |= other
        return FrozenDict(new)

    def __hash__(self) -> int:
        return reduce(lambda acc, pair: acc ^ hash(pair), self.items(), 0)

    def __repr__(self) -> str:
        return repr(self._d)



from collections.abc import Callable
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from edgeql_qb.types import GenericHolder

if TYPE_CHECKING:
    from edgeql_qb.expression import SubQuery  # pragma: no cover


OpLiterals = Literal[
    '=', ':=', '!=', '>', '>=', '<', '<=', '*', '/', '//', '%', '^',
    'and', 'or', '+', '++', '-', 'like', 'ilike', 'in', 'not in', '??', '?=', '?!=',
    'not ', 'exists ', 'not exists ',
]
SortOpLiterals = Literal['asc', 'desc']
Ops: set[OpLiterals] = {
    '=', ':=', '!=', '>', '>=', '<', '<=',
    '*', '/', '//', '%', '^', 'and', 'or', '+', '++', '-',
    'like', 'ilike', 'in', 'not in', '??', '?=', '?!=',
    'not ', 'exists ', 'not exists ',
}
prec_dict: dict[OpLiterals, int] = {
    '^': 11,

    '??': 10,

    '/': 9,
    '*': 9,
    '//': 9,
    '%': 9,

    '+': 8,
    '-': 8,
    '++': 8,

    'in': 7,
    'not in': 7,

    'like': 6,
    'ilike': 6,

    '>': 5,
    '<': 5,
    '>=': 5,
    '<=': 5,

    '=': 4,
    '!=': 4,
    '?=': 4,
    '?!=': 4,

    'not ': 3,
    'exists ': 3,
    'not exists ': 3,

    'and': 2,
    'or': 1,

    ':=': 0,
}
right_assoc_operations = {'^'}
sort_ops = {'asc', 'desc'}
prec_dict_limit = max(prec_dict.values()) + 1


class OperationsMixin:
    def asc(self) -> 'SortedExpression':
        return SortedExpression(self, 'asc')

    def desc(self) -> 'SortedExpression':
        return SortedExpression(self, 'desc')

    def op(self, operation: OpLiterals) -> Callable[[Any], 'BinaryOp']:
        def inner(other: Any) -> BinaryOp:
            return BinaryOp(operation, self, other)
        return inner

    def label(self, name: str) -> 'BinaryOp':
        return BinaryOp(':=', Alias(name), self)

    def like(self, other: Any) -> 'BinaryOp':
        return BinaryOp('like', self, other)

    def ilike(self, other: Any) -> 'BinaryOp':
        return BinaryOp('ilike', self, other)

    def in_(self, other: Any) -> 'BinaryOp':
        return BinaryOp('in', self, other)

    def not_in(self, other: Any) -> 'BinaryOp':
        return BinaryOp('not in', self, other)

    def coalesce(self, other: Any) -> 'BinaryOp':
        return BinaryOp('??', self, other)

    def concat(self, other: Any) -> 'BinaryOp':
        return BinaryOp('++', self, other)

    def exists(self) -> 'UnaryOp':
        return UnaryOp('exists ', self)

    def not_exists(self) -> 'UnaryOp':
        return UnaryOp('not exists ', self)

    def __pos__(self) -> 'UnaryOp':
        return UnaryOp('+', self)

    def __neg__(self) -> 'UnaryOp':
        return UnaryOp('-', self)

    def __invert__(self) -> 'UnaryOp':
        return UnaryOp('not ', self)

    def __and__(self, other: Any) -> 'BinaryOp':
        return BinaryOp('and', self, other)

    def __or__(self, other: Any) -> 'BinaryOp':
        return BinaryOp('or', self, other)

    def __gt__(self, other: Any) -> 'BinaryOp':
        return BinaryOp('>', self, other)

    def __ge__(self, other: Any) -> 'BinaryOp':
        return BinaryOp('>=', self, other)

    def __lt__(self, other: Any) -> 'BinaryOp':
        return BinaryOp('<', self, other)

    def __le__(self, other: Any) -> 'BinaryOp':
        return BinaryOp('<=', self, other)

    def __add__(self, other: Any) -> 'BinaryOp':
        return BinaryOp('+', self, other)

    def __sub__(self, other: Any) -> 'BinaryOp':
        return BinaryOp('-', self, other)

    def __mul__(self, other: Any) -> 'BinaryOp':
        return BinaryOp('*', self, other)

    def __truediv__(self, other: Any) -> 'BinaryOp':
        return BinaryOp('/', self, other)

    def __floordiv__(self, other: Any) -> 'BinaryOp':
        return BinaryOp('//', self, other)

    def __mod__(self, other: Any) -> 'BinaryOp':
        return BinaryOp('%', self, other)


@dataclass(slots=True, frozen=True)
class BinaryOp(OperationsMixin):
    operation: OpLiterals
    left: OperationsMixin
    right: Union[OperationsMixin, 'SubQuery', GenericHolder[Any]]

    def __eq__(self, other: Any) -> 'BinaryOp':  # type: ignore[override]
        return BinaryOp('=', self, other)


@dataclass(slots=True, frozen=True)
class UnaryOp(OperationsMixin):
    operation: OpLiterals
    element: Any


@dataclass(slots=True, frozen=True)
class Alias(OperationsMixin):
    name: str

    def assign(self, value: Union[OperationsMixin, 'SubQuery', GenericHolder[Any]]) -> BinaryOp:
        return BinaryOp(':=', self, value)


@dataclass(slots=True, frozen=True)
class SortedExpression:
    expression: OperationsMixin
    order: Literal['asc', 'desc']


@dataclass(slots=True, frozen=True)
class Node:
    left: 'Node'
    op: OpLiterals
    right: Optional['Node'] = None

    @property
    def precedence(self) -> int:
        return prec_dict[self.op]

    @property
    def assocright(self) -> bool:
        return self.op in right_assoc_operations

    def __repr__(self) -> str:
        return f'Node({self.left!r},{self.op!r},{self.right!r})'

    def __lt__(self, other: Any) -> bool:
        if isinstance(other, Node):
            return self.precedence < other.precedence
        return self.precedence < prec_dict.get(other, prec_dict_limit)

    def __gt__(self, other: Any) -> bool:
        if isinstance(other, Node):
            return self.precedence > other.precedence
        return self.precedence > prec_dict.get(other, prec_dict_limit)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Node):
            return self.precedence == other.precedence
        return self.precedence == prec_dict.get(other, prec_dict_limit)



from collections.abc import Iterator
from functools import reduce, singledispatch
from typing import Any

from edgeql_qb.expression import (
    AnyExpression,
    BaseModel,
    Column,
    Expression,
    QueryLiteral,
    SubQuery,
    UnlessConflict,
    UpdateSubQuery,
)
from edgeql_qb.func import FuncInvocation
from edgeql_qb.operators import Alias, Node
from edgeql_qb.render.func import render_function, render_function_args
from edgeql_qb.render.query_literal import render_query_literal
from edgeql_qb.render.tools import (
    combine_many_renderers,
    join_renderers,
    render_binary_node,
    render_parentheses,
)
from edgeql_qb.render.types import RenderedQuery


def render_insert(model_name: str) -> RenderedQuery:
    return RenderedQuery(f'insert {model_name}')


def render_values(values: list[Expression], generator: Iterator[int]) -> RenderedQuery:
    assert values
    renderers = [
        render_insert_expression(value.to_infix_notation(), generator)
        for value in values
    ]
    return combine_many_renderers(
        RenderedQuery(' { '),
        reduce(join_renderers(', '), renderers),
        RenderedQuery(' }'),
    )


@singledispatch
def render_insert_expression(expression: AnyExpression, generator: Iterator[int]) -> RenderedQuery:
    raise NotImplementedError(f'{expression!r} is not supported')  # pragma: no cover


@render_insert_expression.register
def _(expression: FuncInvocation, generator: Iterator[int]) -> RenderedQuery:
    func = expression.func
    arg_renderers = [
        render_insert_expression(arg, generator)
        for arg in expression.args
    ]
    return render_function(func, arg_renderers)


@render_insert_expression.register
def _(expression: Column, generator: Iterator[int]) -> RenderedQuery:
    return RenderedQuery(expression.column_name)


@render_insert_expression.register
def _(expression: Node, generator: Iterator[int]) -> RenderedQuery:
    assert expression.right is not None, 'Unary operations is not supported in insert expressions'
    return render_binary_node(
        left=render_insert_expression(expression.left, generator),
        right=render_insert_expression(expression.right, generator),
        expression=expression,
    )


@render_insert_expression.register
def _(expression: Alias, generator: Iterator[int]) -> RenderedQuery:
    return RenderedQuery(expression.name)


@render_insert_expression.register
def _(expression: QueryLiteral, generator: Iterator[int]) -> RenderedQuery:
    index = next(generator)
    name = f'insert_{index}'
    return render_query_literal(expression.value, name)


@render_insert_expression.register
def _(expression: SubQuery, generator: Iterator[int]) -> RenderedQuery:
    return expression.build(generator)


@singledispatch
def render_unless_conflict_on(on: Any, generator: Iterator[int]) -> RenderedQuery:
    raise NotImplementedError(f'{on!r} is not supported')  # pragma: no cover


@render_unless_conflict_on.register
def _(on: Column, generator: Iterator[int]) -> RenderedQuery:
    return RenderedQuery(f'.{on.column_name}')


@render_unless_conflict_on.register
def _(
    on: tuple,  # type: ignore[type-arg]
    generator: Iterator[int],
) -> RenderedQuery:
    renderers = [
        render_unless_conflict_on(value, generator)
        for value in on
    ]
    return render_function_args(renderers)


@singledispatch
def render_unless_conflict(conflict: Any, generator: Iterator[int]) -> RenderedQuery:
    raise NotImplementedError(f'{conflict!r} is not supported')  # pragma: no cover


@render_unless_conflict.register
def _(conflict: None, generator: Iterator[int]) -> RenderedQuery:
    return RenderedQuery()


@render_unless_conflict.register
def _(conflict: UnlessConflict, generator: Iterator[int]) -> RenderedQuery:
    rendered_on = (
        combine_many_renderers(
            RenderedQuery(' on '),
            render_unless_conflict_on(conflict.on, generator),
        )
        if conflict.on
        else RenderedQuery()
    )
    rendered_else = (
        combine_many_renderers(
            RenderedQuery(' else '),
            render_unless_conflict_else(conflict.else_, generator),
        )
        if conflict.else_
        else RenderedQuery()
    )
    return combine_many_renderers(
        RenderedQuery(' unless conflict'),
        rendered_on,
        rendered_else,
    )


@singledispatch
def render_unless_conflict_else(else_: Any, generator: Iterator[int]) -> RenderedQuery:
    raise NotImplementedError(f'{else_!r} is not supported')  # pragma: no cover


@render_unless_conflict_else.register
def _(else_: BaseModel, generator: Iterator[int]) -> RenderedQuery:
    return RenderedQuery(else_.name)


@render_unless_conflict_else.register
def _(expression: UpdateSubQuery, generator: Iterator[int]) -> RenderedQuery:
    return render_parentheses(expression.build(generator))



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from edgeql_qb.operators import OperationsMixin


@dataclass(slots=True, frozen=True)
class Function:
    module: str
    name: str

    def __call__(self, *args: Any, **kwargs: Any) -> FuncInvocation:
        return FuncInvocation(func=self, args=args, arity=len(args))


@dataclass(slots=True, frozen=True)
class _Math:
    abs = Function(module='math', name='abs')
    ceil = Function(module='math', name='ceil')
    floor = Function(module='math', name='floor')
    ln = Function(module='math', name='log')
    mean = Function(module='math', name='mean')
    stddev = Function(module='math', name='stddev')
    stddev_pop = Function(module='math', name='stddev_pop')
    var = Function(module='math', name='var')
    var_pop = Function(module='math', name='var_pop')


@dataclass(slots=True, frozen=True)
class _Std:
    all = Function(module='std', name='all')
    any = Function(module='std', name='any')
    array_agg = Function(module='std', name='array_agg')
    array_fill = Function(module='std', name='array_fill')
    array_get = Function(module='std', name='array_get')
    array_join = Function(module='std', name='array_join')
    array_replace = Function(module='std', name='array_replace')
    array_unpack = Function(module='std', name='array_unpack')
    assert_distinct = Function(module='std', name='assert_distinct')
    assert_exists = Function(module='std', name='assert_exists')
    assert_single = Function(module='std', name='assert_single')
    bit_and = Function(module='std', name='bit_and')
    bit_lshift = Function(module='std', name='bit_lshift')
    bit_not = Function(module='std', name='bit_not')
    bit_or = Function(module='std', name='bit_or')
    bit_rshift = Function(module='std', name='bit_rshift')
    bit_xor = Function(module='std', name='bit_xor')
    bytes_get_bit = Function(module='std', name='bytes_get_bit')
    contains = Function(module='std', name='contains')
    count = Function(module='std', name='count')
    datetime_current = Function(module='std', name='datetime_current')
    datetime_get = Function(module='std', name='datetime_get')
    datetime_of_statement = Function(module='std', name='datetime_of_statement')
    datetime_of_transaction = Function(module='std', name='datetime_of_transaction')
    datetime_truncate = Function(module='std', name='datetime_truncate')
    duration_get = Function(module='std', name='duration_get')
    duration_truncate = Function(module='std', name='duration_truncate')
    enumerate = Function(module='std', name='enumerate')
    find = Function(module='std', name='find')
    len = Function(module='std', name='len')
    max = Function(module='std', name='max')
    min = Function(module='std', name='min')
    overlaps = Function(module='std', name='overlaps')
    random = Function(module='std', name='random')
    range_get_lower = Function(module='std', name='range_get_lower')
    range_get_upper = Function(module='std', name='range_get_upper')
    range_is_empty = Function(module='std', name='range_is_empty')
    range_is_inclusive_lower = Function(module='std', name='range_is_inclusive_lower')
    range_is_inclusive_upper = Function(module='std', name='range_is_inclusive_upper')
    range_unpack = Function(module='std', name='range_unpack')
    re_match = Function(module='std', name='re_match')
    re_match_all = Function(module='std', name='re_match_all')
    re_replace = Function(module='std', name='re_replace')
    re_test = Function(module='std', name='re_test')
    sequence_next = Function(module='std', name='sequence_next')
    sequence_reset = Function(module='std', name='sequence_reset')
    str_lower = Function(module='std', name='str_lower')
    str_pad_end = Function(module='std', name='str_pad_end')
    str_pad_start = Function(module='std', name='str_pad_start')
    str_repeat = Function(module='std', name='str_repeat')
    str_replace = Function(module='std', name='str_replace')
    str_reverse = Function(module='std', name='str_reverse')
    str_split = Function(module='std', name='str_split')
    str_title = Function(module='std', name='str_title')
    str_trim = Function(module='std', name='str_trim')
    str_trim_end = Function(module='std', name='str_trim_end')
    str_trim_start = Function(module='std', name='str_trim_start')
    str_upper = Function(module='std', name='str_upper')
    sum = Function(module='std', name='sum')
    to_bigint = Function(module='std', name='to_bigint')
    to_datetime = Function(module='std', name='to_datetime')
    to_decimal = Function(module='std', name='to_decimal')
    to_duration = Function(module='std', name='to_duration')
    to_float32 = Function(module='std', name='to_float32')
    to_float64 = Function(module='std', name='to_float64')
    to_int16 = Function(module='std', name='to_int16')
    to_int32 = Function(module='std', name='to_int32')
    to_int64 = Function(module='std', name='to_int64')
    to_str = Function(module='std', name='to_str')
    uuid_generate_v1mc = Function(module='std', name='uuid_generate_v1mc')
    uuid_generate_v4 = Function(module='std', name='uuid_generate_v4')


@dataclass(slots=True, frozen=True)
class _Sys:
    get_version = Function(module='sys', name='get_version')
    get_version_as_str = Function(module='sys', name='get_version_as_str')
    get_current_database = Function(module='sys', name='get_current_database')


@dataclass(slots=True, frozen=True)
class _Cal:
    local_datetime = Function(module='std', name='local_datetime')
    local_date = Function(module='std', name='local_date')
    local_time = Function(module='std', name='local_time')
    relative_duration = Function(module='std', name='relative_duration')
    date_duration = Function(module='std', name='date_duration')
    to_local_datetime = Function(module='std', name='to_local_datetime')
    to_local_date = Function(module='std', name='to_local_date')
    to_local_time = Function(module='std', name='to_local_time')
    to_relative_duration = Function(module='std', name='to_relative_duration')
    to_date_duration = Function(module='std', name='to_date_duration')
    time_get = Function(module='std', name='time_get')
    date_get = Function(module='std', name='date_get')
    duration_normalize_hours = Function(module='std', name='duration_normalize_hours')
    duration_normalize_days = Function(module='std', name='duration_normalize_days')


math = _Math()
std = _Std()
sys = _Sys()
cal = _Cal()


@dataclass(slots=True, frozen=True)
class FuncInvocation(OperationsMixin):
    func: Function
    args: tuple[Any, ...]
    arity: int



from collections.abc import Iterator
from functools import reduce, singledispatch

from edgeql_qb.expression import (
    AnyExpression,
    Column,
    Expression,
    QueryLiteral,
    SubQuery,
)
from edgeql_qb.func import FuncInvocation
from edgeql_qb.operators import Alias, Node
from edgeql_qb.render.func import render_function
from edgeql_qb.render.query_literal import render_query_literal
from edgeql_qb.render.tools import (
    combine_many_renderers,
    join_renderers,
    render_binary_node,
)
from edgeql_qb.render.types import RenderedQuery


def render_update(model_name: str) -> RenderedQuery:
    return RenderedQuery(f'update {model_name}')


def render_values(values: tuple[Expression, ...], generator: Iterator[int]) -> RenderedQuery:
    assert values
    renderers = [
        render_update_expression(value.to_infix_notation(), generator, '.')
        for value in values
    ]
    return combine_many_renderers(
        RenderedQuery(' set { '),
        reduce(join_renderers(', '), renderers),
        RenderedQuery(' }'),
    )


@singledispatch
def render_update_expression(
        expression: AnyExpression,
        generator: Iterator[int],
        column_prefix: str = '',
) -> RenderedQuery:
    raise NotImplementedError(f'{expression!r} is not supported')  # pragma: no cover


@render_update_expression.register
def _(expression: Column, generator: Iterator[int], column_prefix: str = '') -> RenderedQuery:
    return RenderedQuery(f'{column_prefix}{expression.column_name}')


@render_update_expression.register
def _(expression: Alias, generator: Iterator[int], column_prefix: str = '') -> RenderedQuery:
    return RenderedQuery(expression.name)


@render_update_expression.register
def _(expression: Node, generator: Iterator[int], column_prefix: str = '') -> RenderedQuery:
    assert expression.right is not None, 'Unary operations is not supported in update expressions'
    return render_binary_node(
        left=render_update_expression(
            expression.left,
            generator,
            '.' if expression.op != ':=' else '',
        ),
        right=render_update_expression(expression.right, generator, '.'),
        expression=expression,
    )


@render_update_expression.register
def _(
        expression: FuncInvocation,
        generator: Iterator[int],
        column_prefix: str = '',
) -> RenderedQuery:
    func = expression.func
    arg_renderers = [
        render_update_expression(arg, generator, column_prefix)
        for arg in expression.args
    ]
    return render_function(func, arg_renderers)


@render_update_expression.register
def _(
        expression: QueryLiteral,
        generator: Iterator[int],
        column_prefix: str = '',
) -> RenderedQuery:
    index = next(generator)
    name = f'update_{index}'
    return render_query_literal(expression.value, name)


@render_update_expression.register
def _(expression: SubQuery, generator: Iterator[int], column_prefix: str = '') -> RenderedQuery:
    return expression.build(generator)



from collections.abc import Callable
from dataclasses import dataclass

from edgeql_qb.frozendict import FrozenDict


@dataclass(slots=True, frozen=True)
class RenderedQuery:
    query: str = ''
    context: FrozenDict = FrozenDict()  # noqa: RUF009

    def map(self, f: Callable[['RenderedQuery'], 'RenderedQuery']) -> 'RenderedQuery':
        return f(self)



from functools import reduce

from edgeql_qb.func import Function
from edgeql_qb.render.tools import (
    combine_many_renderers,
    join_renderers,
    render_parentheses,
)
from edgeql_qb.render.types import RenderedQuery


def render_function_args(args: list[RenderedQuery]) -> RenderedQuery:
    return render_parentheses(reduce(join_renderers(', '), args) if args else RenderedQuery())


def render_function(func: Function, arg_renderers: list[RenderedQuery]) -> RenderedQuery:
    return combine_many_renderers(
        RenderedQuery(f'{func.module}::' if func.module != 'std' else ''),
        RenderedQuery(func.name),
        render_function_args(arg_renderers),
    )



from collections.abc import Iterator
from functools import reduce, singledispatch

from edgeql_qb.expression import (
    AnyExpression,
    Column,
    Expression,
    QueryLiteral,
)
from edgeql_qb.func import FuncInvocation
from edgeql_qb.operators import Node, SortedExpression
from edgeql_qb.render.func import render_function
from edgeql_qb.render.query_literal import render_query_literal
from edgeql_qb.render.tools import (
    combine_many_renderers,
    combine_renderers,
    join_renderers,
    render_binary_node,
)
from edgeql_qb.render.types import RenderedQuery


def render_order_by_expressions(
    ordered_by: tuple[Expression, ...],
    generator: Iterator[int],
) -> RenderedQuery:
    renderers = [
        render_order_by_expression(expression.to_infix_notation(), generator)
        for expression in ordered_by
    ]
    return combine_renderers(
        RenderedQuery(' order by '),
        reduce(join_renderers(' then '), renderers),
    )


def render_order_by(ordered_by: tuple[Expression, ...], generator: Iterator[int]) -> RenderedQuery:
    return render_order_by_expressions(ordered_by, generator) if ordered_by else RenderedQuery()


@singledispatch
def render_order_by_expression(
        expression: AnyExpression,
        generator: Iterator[int],
) -> RenderedQuery:
    raise NotImplementedError(f'{expression!r} is not supported')  # pragma: no cover


@render_order_by_expression.register
def _(expression: Column, generator: Iterator[int]) -> RenderedQuery:
    return RenderedQuery(f'.{expression.column_name}')


@render_order_by_expression.register
def _(expression: FuncInvocation, generator: Iterator[int]) -> RenderedQuery:
    func = expression.func
    arg_renderers = [
        render_order_by_expression(arg, generator)
        for arg in expression.args
    ]
    return render_function(func, arg_renderers)


@render_order_by_expression.register
def _(expression: Node, generator: Iterator[int]) -> RenderedQuery:
    if expression.right is None:
        return combine_many_renderers(
            RenderedQuery(expression.op),
            render_order_by_expression(expression.left, generator),
        )
    return render_binary_node(
        left=render_order_by_expression(expression.left, generator),
        right=render_order_by_expression(expression.right, generator),
        expression=expression,
    )


@render_order_by_expression.register
def _(expression: SortedExpression, generator: Iterator[int]) -> RenderedQuery:
    return combine_renderers(
        render_order_by_expression(expression.expression, generator),
        RenderedQuery(f' {expression.order}'),
    )


@render_order_by_expression.register
def _(expression: QueryLiteral, generator: Iterator[int]) -> RenderedQuery:
    index = next(generator)
    name = f'order_by_{index}'
    return render_query_literal(expression.value, name)



from collections.abc import Callable
from functools import reduce
from typing import cast

from edgeql_qb.expression import Column, SubQuery
from edgeql_qb.operators import Node
from edgeql_qb.render.types import RenderedQuery


def combine_many_renderers(*renderers: RenderedQuery) -> RenderedQuery:
    return reduce(combine_renderers, renderers)


def join_renderers(separator: str = '') -> Callable[[RenderedQuery, RenderedQuery], RenderedQuery]:
    def inner(r1: RenderedQuery, r2: RenderedQuery) -> RenderedQuery:
        return RenderedQuery(
            query=f'{r1.query}{separator}{r2.query}',
            context=r1.context | r2.context,
        )
    return inner


def combine_renderers(r1: RenderedQuery, r2: RenderedQuery) -> RenderedQuery:
    return join_renderers()(r1, r2)


def need_left_parentheses(left: Node, expression: Node) -> bool:
    return (
        isinstance(left, Node)
        and left < expression
        or (
            left == expression
            and not expression.assocright
            and (
                not isinstance(left, Column)
                and getattr(left, 'operation', 1) != getattr(expression, 'operation', 2)
            )
        )
    )


def need_right_parentheses(right: Node, expression: Node) -> bool:
    return (
        isinstance(right, Node)
        and right < expression
        or (isinstance(right, SubQuery))
        or (isinstance(right, Node) and right == expression and expression.assocright)
        or (getattr(right, 'op', None) == '-' and expression.assocright)
    )


def render_parentheses(inner: RenderedQuery) -> RenderedQuery:
    return combine_many_renderers(RenderedQuery('('), inner, RenderedQuery(')'))


def render_assoc_parentheses(
        is_need: bool,  # noqa: FBT001
) -> Callable[[RenderedQuery], RenderedQuery]:
    def inner(column: RenderedQuery) -> RenderedQuery:
        return is_need and render_parentheses(column) or column
    return inner


def render_right_parentheses(
    right: Node,
    expression: Node,
) -> Callable[[RenderedQuery], RenderedQuery]:
    is_need = need_right_parentheses(right, expression)
    return render_assoc_parentheses(is_need=is_need)


def render_left_expression(query: RenderedQuery) -> Callable[[Node], RenderedQuery]:
    def inner(expression: Node) -> RenderedQuery:
        left = render_left_parentheses(expression.left, expression)
        return query.map(left)
    return inner


def render_right_expression(query: RenderedQuery) -> Callable[[Node], RenderedQuery]:
    def inner(expression: Node) -> RenderedQuery:
        right = render_right_parentheses(cast(Node, expression.right), expression)
        return query.map(right)
    return inner


def render_binary_node(
    left: RenderedQuery,
    right: RenderedQuery,
    expression: Node,
) -> RenderedQuery:
    left = render_left_expression(left)(expression)
    right = render_right_expression(right)(expression)
    return join_renderers(f' {expression.op} ')(left, right)


def render_left_parentheses(
    left: Node,
    expression: Node,
) -> Callable[[RenderedQuery], RenderedQuery]:
    is_need = need_left_parentheses(left, expression)
    return render_assoc_parentheses(is_need=is_need)


def linearize_filter_left(column: Column) -> list[Column]:
    match column.parent:
        case None:
            return [column]
        case _:
            return [*linearize_filter_left(column.parent), column]



from abc import ABC, abstractmethod
from collections import deque
from collections.abc import Iterator
from dataclasses import dataclass, field, replace
from enum import Enum, auto
from functools import singledispatch
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Optional,
    TypeVar,
    Union,
    cast,
)

from edgeql_qb.func import FuncInvocation
from edgeql_qb.operators import (
    Alias,
    BinaryOp,
    Node,
    OperationsMixin,
    OpLiterals,
    Ops,
    SortedExpression,
    UnaryOp,
    sort_ops,
)
from edgeql_qb.types import GenericHolder, unsafe_text

if TYPE_CHECKING:
    from edgeql_qb.queries import EdgeDBModel  # pragma: no cover
    from edgeql_qb.render.types import RenderedQuery  # pragma: no cover


@dataclass(slots=True, frozen=True)
class QueryLiteral:
    value: Any


class SymbolType(Enum):
    column = auto()
    shape = auto()
    sort_direction = auto()
    literal = auto()
    operator = auto()
    subquery = auto()
    alias = auto()
    text = auto()
    func_invocation = auto()


@dataclass(slots=True, frozen=True)
class Shape:
    parent: 'Column'
    columns: tuple[Union['Column', 'Shape'], ...]
    filters: tuple['Expression', ...] = field(default_factory=tuple)
    ordered_by: tuple['Expression', ...] = field(default_factory=tuple)
    limit_val: int | unsafe_text | FuncInvocation | None = None
    offset_val: int | unsafe_text | FuncInvocation | None = None

    def where(self, compared: Union['BinaryOp', 'UnaryOp', 'FuncInvocation']) -> 'Shape':
        return replace(self, filters=(*self.filters, Expression(compared)))

    def order_by(
        self,
        *columns: Union[SortedExpression, 'Column', UnaryOp, FuncInvocation],
    ) -> 'Shape':
        new_expressions = [Expression(exp) for exp in columns]
        return replace(self, ordered_by=(*self.ordered_by, *new_expressions))

    def limit(self, value: int | FuncInvocation | unsafe_text) -> 'Shape':
        return replace(self, limit_val=value)

    def offset(self, value: int | FuncInvocation | unsafe_text) -> 'Shape':
        return replace(self, offset_val=value)


@dataclass(slots=True, frozen=True)
class Column(OperationsMixin):
    column_name: str
    parent: Optional['Column'] = None

    def __call__(self, *columns: Union['Column', Shape]) -> Shape:
        return Shape(self, columns)

    def __getattr__(self, name: str) -> 'Column':
        return Column(name, self)

    def __eq__(self, other: Any) -> BinaryOp:  # type: ignore[override]
        return BinaryOp('=', self, other)

    def __ne__(self, other: Any) -> BinaryOp:  # type: ignore[override]
        return BinaryOp('!=', self, other)


SelectExpressions = (
    Column
    | Shape
    | BinaryOp
    | Node
    | QueryLiteral
)


@dataclass(slots=True, frozen=True)
class BaseModel:
    name: str
    module: str | None = None
    schema: str = 'default'


class SubQuery(ABC):
    @abstractmethod
    def build(self, generator: Iterator[int] | None = None) -> 'RenderedQuery':
        raise NotImplementedError()  # pragma: no cover

    def label(self, name: str) -> BinaryOp:
        return BinaryOp(':=', Alias(name), self)


class UpdateSubQuery(SubQuery):
    pass


@dataclass(slots=True, frozen=True)
class UnlessConflict:
    on: tuple[Column, ...] | Column | None
    else_: Union[UpdateSubQuery, 'EdgeDBModel', None] = None


AnyExpression = (
    Column
    | Shape
    | BinaryOp
    | Node
    | QueryLiteral
    | UnaryOp
    | SortedExpression
    | OperationsMixin
    | SubQuery
    | FuncInvocation
    | GenericHolder[Any]
    | unsafe_text
)
FilterExpressions = BinaryOp | UnaryOp
StackType = TypeVar('StackType')


class Stack(Generic[StackType]):
    def __init__(self) -> None:
        self._stack = deque[StackType]()

    def push(self, frame: StackType) -> None:
        self._stack.append(frame)

    def pop(self) -> StackType:
        return self._stack.pop()

    def popn(self, argcount: int) -> list[StackType]:
        return [self._stack.pop() for _ in range(argcount)]


@singledispatch
def _determine_type(value: Any) -> SymbolType:
    return SymbolType.literal


@_determine_type.register
def _(value: Column) -> SymbolType:
    return SymbolType.column


@_determine_type.register
def _(value: Shape) -> SymbolType:
    return SymbolType.shape


@_determine_type.register
def _(value: SubQuery) -> SymbolType:
    return SymbolType.subquery


@_determine_type.register
def _(value: Alias) -> SymbolType:
    return SymbolType.alias


@_determine_type.register
def _(value: unsafe_text) -> SymbolType:
    return SymbolType.text


@_determine_type.register
def _(value: FuncInvocation) -> SymbolType:
    return SymbolType.func_invocation


@_determine_type.register
def _(value: str) -> SymbolType:
    if value in sort_ops:
        return SymbolType.sort_direction
    elif value in Ops:
        return SymbolType.operator
    else:
        return SymbolType.literal


class Symbol:
    __slots__ = 'value', 'type', 'arity', 'depth'

    def __init__(self, value: Any, arity: int | None = None, depth: int = 0):
        self.value = value
        self.type = _determine_type(value)
        self.arity = arity
        self.depth = depth


def build_binary_op(op: OpLiterals, left: Node, right: Node) -> Node:
    if op == getattr(right, 'op', None) == '-' and right.right is None:
        # equals to: a - -b = a + b
        return Node(left, '+', right.left)
    elif (
        (op == '+' and getattr(right, 'op', None) == '-')
        or (op == '-' and getattr(right, 'op', None) == '+')
        and right.right is None
    ):
        # equals to: a + -b = a - +b = a - b
        return Node(left, '-', right.left)
    return Node(left, op, right)


def build_unary_op(op: OpLiterals, argument: Node) -> Node:
    return Node(argument, op)


def evaluate(stack: Stack[AnyExpression], symbol: Symbol) -> None:  # noqa: C901
    match symbol.type:
        case (
            SymbolType.column
            | SymbolType.shape
            | SymbolType.text
            | SymbolType.alias
            | SymbolType.subquery
        ):
            stack.push(symbol.value)
        case SymbolType.operator:
            match symbol.arity:
                case 1:
                    argument = stack.pop()
                    stack.push(build_unary_op(symbol.value, cast(Node, argument)))
                case 2:
                    left, right = stack.popn(2)
                    stack.push(build_binary_op(symbol.value, cast(Node, left), cast(Node, right)))
                case _:  # pragma: no cover
                    assert False
        case SymbolType.literal:
            stack.push(QueryLiteral(symbol.value))
        case SymbolType.sort_direction:
            sorted_expression = stack.pop()
            stack.push(SortedExpression(cast(OperationsMixin, sorted_expression), symbol.value))
        case SymbolType.func_invocation:
            args = stack.popn(symbol.value.arity)
            invocation = FuncInvocation(
                func=symbol.value.func,
                args=tuple(args),
                arity=symbol.value.arity,
            )
            stack.push(invocation)
        case _:  # pragma: no cover
            assert False


def _replace_alias_with_label(node: Any, depth: int) -> Any:
    """Replace assignment operation with label.

    Top level expression should not be replaced, so depth checking is necessary as well.
    Node(':=', left=Alias('test'), right=1) -> Alias('test').
    """
    match node:
        case BinaryOp(':=', left, _) if depth > 0:
            return left
        case _:
            return node


class Expression:
    def __init__(self, expression: AnyExpression):
        self.serialized = tuple(self._to_polish_notation(expression))

    def to_infix_notation(self) -> 'AnyExpression':
        stack = Stack[AnyExpression]()
        for symbol in reversed(self.serialized):
            evaluate(stack, symbol)
        return stack.pop()

    def _to_polish_notation(  # noqa: C901
            self,
            expr: AnyExpression,
            depth: int = 0,
    ) -> Iterator[Symbol]:
        match expr:
            case Column() | Alias():
                yield Symbol(expr, depth=depth)
            case UnaryOp(operation, element):
                yield Symbol(operation, arity=1, depth=depth)
                # a := -(b := 1) -> a := -b
                element = _replace_alias_with_label(element, depth)
                yield from self._to_polish_notation(element, depth + 1)
            case BinaryOp(operation, left, right):
                yield Symbol(operation, arity=2, depth=depth)

                new_depth = depth + 1
                # a := (b := value) + 1 -> a := b + 1
                left = _replace_alias_with_label(left, new_depth)
                # a := 1 + (b := value) -> a := 1 + b
                right = _replace_alias_with_label(right, new_depth)

                yield from self._to_polish_notation(left, new_depth)
                yield from self._to_polish_notation(right, new_depth)
            case SortedExpression(expression, direction):
                yield Symbol(direction, depth=depth)
                yield from self._to_polish_notation(expression, depth + 1)
            case FuncInvocation(_, args):
                yield Symbol(expr, arity=expr.arity, depth=depth)
                for arg in args:
                    # a := fun(b := 1, c := 2) -> a := fun(b, c)
                    flat_arg = _replace_alias_with_label(arg, depth)
                    yield from self._to_polish_notation(flat_arg, depth + 1)
            case _:
                yield Symbol(expr, depth=depth)


class Columns:
    def __getattribute__(self, name: str) -> Column:
        return Column(name)



from collections.abc import Iterator
from functools import reduce, singledispatch

from edgeql_qb.expression import (
    AnyExpression,
    Column,
    Expression,
    QueryLiteral,
    Shape,
    SubQuery,
)
from edgeql_qb.func import FuncInvocation
from edgeql_qb.operators import Alias, Node
from edgeql_qb.render.condition import render_conditions
from edgeql_qb.render.func import render_function
from edgeql_qb.render.order_by import render_order_by
from edgeql_qb.render.pagination import render_limit, render_offset
from edgeql_qb.render.query_literal import render_query_literal
from edgeql_qb.render.tools import (
    combine_many_renderers,
    combine_renderers,
    join_renderers,
    render_binary_node,
    render_parentheses,
)
from edgeql_qb.render.types import RenderedQuery


def render_select_columns(
        select: tuple[Expression, ...],
        generator: Iterator[int],
) -> RenderedQuery:
    renderers = (
        render_select_expression(selectable.to_infix_notation(), generator)
        for selectable in select
    )
    return combine_many_renderers(
        RenderedQuery(' { '),
        reduce(join_renderers(', '), renderers),
        RenderedQuery(' }'),
    )


def render_select(
    model_name: str,
    select: tuple[Expression, ...],
    generator: Iterator[int],
    select_from_query: SubQuery | None = None,
) -> RenderedQuery:
    rendered_select = (
        render_parentheses(select_from_query.build(generator))
        if select_from_query
        else RenderedQuery(model_name)
    )
    return combine_many_renderers(
        RenderedQuery('select '),
        rendered_select,
    ).map(
        lambda r: (
            combine_renderers(r, render_select_columns(select, generator))
            if select
            else r
        ),
    )


@singledispatch
def render_select_expression(
        expression: AnyExpression,
        generator: Iterator[int],
        column_prefix: str = '',
) -> RenderedQuery:
    raise NotImplementedError(f'{expression!r} is not supported')  # pragma: no cover


@render_select_expression.register
def _(expression: Column, generator: Iterator[int], column_prefix: str = '') -> RenderedQuery:
    return RenderedQuery(f'{column_prefix}{expression.column_name}')


@render_select_expression.register
def _(
        expression: FuncInvocation,
        generator: Iterator[int],
        column_prefix: str = '',
) -> RenderedQuery:
    func = expression.func
    arg_renderers = [
        render_select_expression(arg, generator, column_prefix)
        for arg in expression.args
    ]
    return render_function(func, arg_renderers)


@render_select_expression.register
def _(expression: Alias, generator: Iterator[int], column_prefix: str = '') -> RenderedQuery:
    return RenderedQuery(expression.name)


@render_select_expression.register
def _(expression: Shape, generator: Iterator[int], column_prefix: str = '') -> RenderedQuery:
    expressions = (
        render_select_expression(exp, generator, column_prefix)
        for exp in expression.columns
    )
    conditions = render_conditions(expression.filters, generator=generator)
    order_by = render_order_by(expression.ordered_by, generator=generator)
    rendered_offset = render_offset(expression.offset_val, generator=generator)
    rendered_limit = render_limit(expression.limit_val, generator=generator)
    return combine_many_renderers(
        RenderedQuery(f'{expression.parent.column_name}: {{ '),
        reduce(join_renderers(', '), expressions),
        RenderedQuery(' }'),
        conditions,
        order_by,
        rendered_offset,
        rendered_limit,
    )


@render_select_expression.register
def _(
        expression: QueryLiteral,
        generator: Iterator[int],
        column_prefix: str = '',
) -> RenderedQuery:
    index = next(generator)
    name = f'select_{index}'
    return render_query_literal(expression.value, name)


@render_select_expression.register
def _(expression: Node, generator: Iterator[int], column_prefix: str = '') -> RenderedQuery:
    if expression.right is None:
        return combine_many_renderers(
            RenderedQuery(expression.op),
            render_select_expression(expression.left, generator, column_prefix),
        )
    return render_binary_node(
        left=render_select_expression(
            expression.left,
            generator,
            expression.op != ':=' and '.' or '',
        ),
        right=render_select_expression(expression.right, generator, '.'),
        expression=expression,
    )


@render_select_expression.register
def _(expression: SubQuery, generator: Iterator[int], column_prefix: str = '') -> RenderedQuery:
    return expression.build(generator)



from collections.abc import Iterator
from functools import singledispatch
from typing import Any

from edgeql_qb.expression import Expression, QueryLiteral
from edgeql_qb.func import FuncInvocation
from edgeql_qb.render.func import render_function
from edgeql_qb.render.query_literal import render_query_literal
from edgeql_qb.render.tools import combine_many_renderers
from edgeql_qb.render.types import RenderedQuery
from edgeql_qb.types import int64, unsafe_text


@singledispatch
def render_offset(offset: Any, generator: Iterator[int]) -> RenderedQuery:
    raise NotImplementedError(f'{offset!r} is not supported')  # pragma: no cover


@render_offset.register
def _(offset: None, generator: Iterator[int]) -> RenderedQuery:
    return RenderedQuery()


@render_offset.register
def _(offset: QueryLiteral, generator: Iterator[int]) -> RenderedQuery:
    index = next(generator)
    name = f'offset_{index}'
    return render_query_literal(offset.value, name)


@render_offset.register
def _(offset: FuncInvocation, generator: Iterator[int]) -> RenderedQuery:
    func = offset.func
    arg_renderers = [
        render_offset(Expression(arg).to_infix_notation(), generator)
        for arg in offset.args
    ]
    return combine_many_renderers(
        RenderedQuery(' offset '),
        render_function(func, arg_renderers),
    )


@render_offset.register
def _(offset: int, generator: Iterator[int]) -> RenderedQuery:
    index = next(generator)
    name = f'offset_{index}'
    return combine_many_renderers(
        RenderedQuery(' offset '),
        render_query_literal(int64(offset), name),
    )


@render_offset.register
def _(offset: unsafe_text, generator: Iterator[int]) -> RenderedQuery:
    return RenderedQuery(f' offset {offset!s}')


@singledispatch
def render_limit(limit: Any, generator: Iterator[int]) -> RenderedQuery:
    raise NotImplementedError(f'{limit!r} is not supported')  # pragma: no cover


@render_limit.register
def _(limit: None, generator: Iterator[int]) -> RenderedQuery:
    return RenderedQuery()


@render_limit.register
def _(limit: int, generator: Iterator[int]) -> RenderedQuery:
    index = next(generator)
    name = f'limit_{index}'
    return combine_many_renderers(
        RenderedQuery(' limit '),
        render_query_literal(int64(limit), name),
    )


@render_limit.register
def _(limit: unsafe_text, generator: Iterator[int]) -> RenderedQuery:
    return RenderedQuery(f' limit {limit!s}')


@render_limit.register
def _(limit: QueryLiteral, generator: Iterator[int]) -> RenderedQuery:
    index = next(generator)
    name = f'limit_{index}'
    return render_query_literal(limit.value, name)


@render_limit.register
def _(limit: FuncInvocation, generator: Iterator[int]) -> RenderedQuery:
    func = limit.func
    arg_renderers = [
        render_limit(Expression(arg).to_infix_notation(), generator)
        for arg in limit.args
    ]
    return combine_many_renderers(
        RenderedQuery(' limit '),
        render_function(func, arg_renderers),
    )



from collections.abc import Iterator
from functools import singledispatch

from edgeql_qb.expression import (
    AnyExpression,
    Column,
    Expression,
    QueryLiteral,
    SubQuery,
)
from edgeql_qb.func import FuncInvocation
from edgeql_qb.operators import Alias, Node
from edgeql_qb.render.func import render_function
from edgeql_qb.render.query_literal import render_query_literal
from edgeql_qb.render.tools import combine_many_renderers, render_binary_node
from edgeql_qb.render.types import RenderedQuery


@singledispatch
def render_expression(
    expression: AnyExpression,
    literal_prefix: str,
    generator: Iterator[int],
    column_prefix: str = '',
) -> RenderedQuery:
    raise NotImplementedError(f'{expression!r} is not supported')  # pragma: no cover


@render_expression.register
def _(
    expression: SubQuery,
    literal_prefix: str,
    generator: Iterator[int],
    column_prefix: str = '',
) -> RenderedQuery:
    return expression.build(generator)


@render_expression.register
def _(
    expression: FuncInvocation,
    literal_prefix: str,
    generator: Iterator[int],
    column_prefix: str = '',
) -> RenderedQuery:
    func = expression.func
    arg_renderers = [
        render_expression(
            Expression(arg).to_infix_notation(),
            literal_prefix,
            generator,
            column_prefix,
        )
        for arg in expression.args
    ]
    return render_function(func, arg_renderers)


@render_expression.register
def _(
        expression: Column,
        literal_prefix: str,
        generator: Iterator[int],
        column_prefix: str = '',
) -> RenderedQuery:
    return RenderedQuery(f'{column_prefix}{expression.column_name}')


@render_expression.register
def _(
        expression: Alias,
        literal_prefix: str,
        generator: Iterator[int],
        column_prefix: str = '',
) -> RenderedQuery:
    return RenderedQuery(expression.name)


@render_expression.register
def _(
    expression: QueryLiteral,
    literal_prefix: str,
    generator: Iterator[int],
    column_prefix: str = '',
) -> RenderedQuery:
    index = next(generator)
    name = f'{literal_prefix}_{index}'
    return render_query_literal(expression.value, name)


@render_expression.register
def _(
    expression: Node,
    literal_prefix: str,
    generator: Iterator[int],
    column_prefix: str = '',
) -> RenderedQuery:
    if expression.right is None:
        return combine_many_renderers(
            RenderedQuery(expression.op),
            render_expression(expression.left, literal_prefix, generator, column_prefix),
        )
    return render_binary_node(
        left=render_expression(
            expression.left,
            literal_prefix,
            generator,
            column_prefix=expression.op != ':=' and '.' or '',
        ),
        right=render_expression(expression.right, literal_prefix, generator, '.'),
        expression=expression,
    )
